# -*- coding: utf-8 -*-
"""
Phlo client, used for Phlo API requests.
"""
from plivo.resources import Phlos
from plivo.rest.base_client import BaseClient
from requests import Request

PHLO_API = 'https://phlorunner.plivo.com'
PHLO_API_BASE_URI = '/'.join([PHLO_API, 'v1'])


class PhloClient(BaseClient):
    def __init__(self, auth_id=None, auth_token=None, proxies=None, timeout=5):
        """
        The Plivo API client.

        Deals with all the API requests to be made.
        """
        BaseClient.__init__(self, auth_id, auth_token, proxies, timeout)

        self.phlo_base_uri = PHLO_API_BASE_URI
        self.phlo = Phlos(self)

    def create_request(self, method, path=None, data=None):
        path = path or []

        req = Request(method, '/'.join([self.phlo_base_uri] +
                                       list([str(p) for p in path])) + '/',
                      **({
                          'params': data
                      } if method == 'GET' else {
                          'json': data
                      }))
        return self.session.prepare_request(req)


# -*- coding: utf-8 -*-
"""
Call class - along with its list class
"""

from plivo.base import (ListResponseObject, PlivoResource,
                        PlivoResourceInterface)
from plivo.utils import to_param_dict
from plivo.utils.validators import *


class Call(PlivoResource):
    _name = 'Call'
    _identifier_string = 'call_uuid'

    def update(self,
               legs=None,
               aleg_url=None,
               aleg_method=None,
               bleg_url=None,
               bleg_method=None):
        return self.client.calls.update(self.id,
                                        **to_param_dict(self.update, locals()))

    def get(self):
        return self.client.calls.get(self.id)

    def record(self,
               time_limit=None,
               file_format=None,
               transcription_type=None,
               transcription_url=None,
               transcription_method=None,
               callback_url=None,
               callback_method=None):
        return self.client.calls.record(self.id,
                                        **to_param_dict(self.record, locals()))

    def start_recording(self,
                        time_limit=None,
                        file_format=None,
                        transcription_type=None,
                        transcription_url=None,
                        transcription_method=None,
                        callback_url=None,
                        callback_method=None):
        return self.client.calls.start_recording(self.id,
                                                 **to_param_dict(
                                                     self.start_recording,
                                                     locals()))

    def stop_recording(self):
        return self.client.calls.stop_recording(self.id)

    def play(self, urls, length=None, legs=None, loop=None, mix=None):
        return self.client.calls.play(self.id,
                                      **to_param_dict(self.start_playing,
                                                      locals()))

    def start_playing(self, urls, length=None, legs=None, loop=None, mix=None):
        return self.client.calls.play(self.id,
                                      **to_param_dict(self.start_playing,
                                                      locals()))

    def stop_playing(self):
        return self.client.calls.stop_playing(self.id)

    def speak(self,
              call_uuid,
              text,
              voice=None,
              language=None,
              legs=None,
              loop=None,
              mix=None,
              type=None, is_voice_request=True):
        return self.client.calls.speak(self.id,
                                       **to_param_dict(self.start_playing,
                                                       locals()))

    def start_speaking(self,
                       text,
                       voice=None,
                       language=None,
                       legs=None,
                       loop=None,
                       mix=None,
                       type=None):
        return self.client.calls.start_speaking(self.id,
                                                **to_param_dict(
                                                    self.start_playing,
                                                    locals()))

    def stop_speaking(self):
        return self.client.calls.stop_speaking(self.id)

    def send_digits(self, digits, leg):
        return self.client.calls.send_digits(self.id,
                                             **to_param_dict(
                                                 self.start_playing, locals()))

    def delete(self):
        return self.client.calls.delete(self.id)

    def hangup(self):
        return self.client.calls.delete(self.id)

    def start_stream(self,
                     service_url,
                     bidirectional=None,
                     audio_track=None,
                     stream_timeout=None,
                     status_callback_url=None,
                     status_callback_method=None,
                     content_type=None,
                     extra_headers=None):
        return self.client.calls.start_stream(self.id,
                                              **to_param_dict(
                                                  self.start_stream(),
                                                  locals()))

    def delete_specific_stream(self,
                               stream_id):
        return self.client.calls.delete_specific_stream(self.id,
                                                        **to_param_dict(
                                                            self.delete_specific_stream(),
                                                            locals()))

    def delete_all_streams(self):
        return self.client.calls.delete_all_streams(self.id,
                                                    **to_param_dict(
                                                        self.delete_all_streams(),
                                                        locals()))

    def get_details_of_specific_stream(self,
                                       stream_id):
        return self.client.calls.get_details_of_specific_stream(self.id,
                                                                **to_param_dict(
                                                                    self.get_details_of_specific_stream(),
                                                                    locals()))

    def get_all_streams(self):
        return self.client.calls.get_all_streams(self.id,
                                                 **to_param_dict(
                                                     self.get_details_of_specific_stream(),
                                                     locals()))


class Calls(PlivoResourceInterface):
    _resource_type = Call

    @validate_args(
        from_=[is_phonenumber()],
        to_=[is_iterable(of_type(six.text_type), sep='<')],
        answer_url=[is_url()],
        answer_method=[optional(of_type(six.text_type))],
        hangup_url=[optional(is_url())],
        hangup_method=[optional(of_type(six.text_type))],
        fallback_url=[optional(is_url())],
        fallback_method=[optional(of_type(six.text_type))],
        ring_url=[optional(is_url())],
        ring_method=[optional(of_type(six.text_type))],
        machine_detection_url=[optional(is_url())],
        machine_detection_method=[optional(of_type(six.text_type))],
        caller_name=[optional(of_type(six.text_type))],
        error_if_parent_not_found=[optional(of_type_exact(bool))],
        parent_call_uuid=[optional(of_type(six.text_type))],
        ring_timeout=[
            optional(of_type(*six.integer_types)),
            check(
                lambda ring_timeout: 0 <= ring_timeout,
                message='0 <= ring_timeout')
        ],
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))],
    )
    def create(self,
               from_,
               to_,
               answer_url,
               answer_method='POST',
               ring_url=None,
               ring_method='POST',
               hangup_url=None,
               hangup_method='POST',
               fallback_url=None,
               fallback_method='POST',
               caller_name=None,
               send_digits=None,
               send_on_preanswer=None,
               time_limit=None,
               hangup_on_ring=None,
               machine_detection=None,
               machine_detection_time=5000,
               machine_detection_url=None,
               machine_detection_method='POST',
               sip_headers=None,
               ring_timeout=120,
               parent_call_uuid=None,
               error_if_parent_not_found=False,
               callback_url=None,
               callback_method=None):
        if from_ in to_.split('<'):
            raise ValidationError('src and destination cannot overlap')
        return self.client.request('POST', ('Call',), to_param_dict(self.create, locals()), is_voice_request=True)

    @validate_args(
        subaccount=[optional(is_subaccount())],
        call_direction=[
            optional(of_type(six.text_type), is_in(('inbound', 'outbound')))
        ],
        from_number=[optional(is_phonenumber())],
        to_number=[optional(is_iterable(of_type(six.text_type), sep='<'))],
        limit=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda limit: 0 < limit <= 20, '0 < limit <= 20')))
        ],
        offset=[
            optional(
                all_of(
                    of_type(*six.integer_types),
                    check(lambda offset: 0 <= offset, '0 <= offset')))
        ],
        parent_call_uuid=[
            optional(of_type(six.text_type))
        ],
        hangup_cause_code=[
            optional(
                all_of(
                    of_type(*six.integer_types)))
        ],
        hangup_source=[
            optional(of_type(six.text_type))
        ],
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))],
    )
    def list(self,
             subaccount=None,
             call_direction=None,
             call_direction__lt=None,
             call_direction__lte=None,
             call_direction__gt=None,
             call_direction__gte=None,
             end_time__lt=None,
             end_time__lte=None,
             end_time__gt=None,
             end_time__gte=None,
             end_time=None,
             from_number=None,
             to_number=None,
             bill_duration=None,
             limit=20,
             offset=0,
             status=None,
             parent_call_uuid=None,
             hangup_cause_code=None,
             hangup_source=None,
             callback_url=None,
             callback_method=None
             ):
        # Adding if else block because if we are fetching response without callback_url then response will be of type
        # ListResponseObject, if passing callback_url then will be of type
        # {'api_id': '94722e88-ae7c-11ec-b52e-0242ac11000a', 'message': 'async api spawned'}
        if callback_url:
            return self.client.request(
                'GET',
                ('Call',),
                to_param_dict(self.list, locals()),
                is_voice_request=True
            )
        else:
            return self.client.request(
                'GET',
                ('Call',),
                to_param_dict(self.list, locals()),
                response_type=ListResponseObject,
                is_voice_request=True
            )

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))],
    )
    def get(self,
            call_uuid,
            callback_url=None,
            callback_method=None):
        return self.client.request('GET', ('Call', call_uuid), to_param_dict(self.get, locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        legs=[of_type(six.text_type),
              is_in(('aleg', 'bleg', 'both'))],
        aleg_url=[optional(is_url())],
        aleg_method=[optional(of_type(six.text_type))],
        bleg_method=[optional(of_type(six.text_type))],
        bleg_url=[optional(is_url())],
    )
    def update(self,
               call_uuid,
               legs=None,
               aleg_url=None,
               aleg_method=None,
               bleg_url=None,
               bleg_method=None,
               callback_url=None,
               callback_method=None):
        return self.client.request('POST', ('Call', call_uuid),
                                   to_param_dict(self.update, locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))],
    )
    def transfer(self,
                 call_uuid,
                 legs=None,
                 aleg_url=None,
                 aleg_method=None,
                 bleg_url=None,
                 bleg_method=None,
                 callback_url=None,
                 callback_method=None):
        return self.update(
            call_uuid,
            legs=legs,
            aleg_method=aleg_method,
            aleg_url=aleg_url,
            bleg_url=bleg_url,
            bleg_method=bleg_method,
            callback_url=callback_url,
            callback_method=callback_method)

    @validate_args(call_uuid=[of_type(six.text_type)])
    def record(self,
               call_uuid,
               time_limit=None,
               file_format=None,
               transcription_type=None,
               transcription_url=None,
               transcription_method=None,
               callback_url=None,
               callback_method=None):
        return self.start_recording(**to_param_dict(self.start_recording,
                                                    locals()))

    @validate_args(call_uuid=[of_type(six.text_type)])
    def start_recording(self,
                        call_uuid,
                        time_limit=None,
                        file_format=None,
                        transcription_type=None,
                        transcription_url=None,
                        transcription_method=None,
                        callback_url=None,
                        callback_method=None):
        return self.client.request('POST', ('Call', call_uuid, 'Record'),
                                   to_param_dict(self.start_recording,
                                                 locals()), is_voice_request=True)

    def record_stop(self, call_uuid, callback_url=None, callback_method=None):
        return self.client.calls.stop_recording(call_uuid, callback_url, callback_method)

    @validate_args(call_uuid=[of_type(six.text_type)],
                   callback_url=[optional(is_url())],
                   callback_method=[optional(of_type(six.text_type))]
                   )
    def stop_recording(self, call_uuid, callback_url, callback_method):
        return self.client.request('DELETE', ('Call', call_uuid, 'Record'),
                                   to_param_dict(self.stop_recording, locals()), is_voice_request=True)

    @validate_args(call_uuid=[of_type(six.text_type)])
    def play(self,
             call_uuid,
             urls,
             length=None,
             legs=None,
             loop=None,
             mix=None,
             callback_url=None,
             callback_method=None):
        return self.start_playing(**to_param_dict(self.play, locals()))

    @validate_args(call_uuid=[of_type(six.text_type)])
    def start_playing(self,
                      call_uuid,
                      urls,
                      length=None,
                      legs=None,
                      loop=None,
                      mix=None,
                      callback_url=None,
                      callback_method=None):
        return self.client.request('POST', ('Call', call_uuid, 'Play'),
                                   to_param_dict(self.play, locals()), is_voice_request=True)

    def play_stop(self, call_uuid, callback_url=None, callback_method=None):
        return self.client.calls.stop_playing(call_uuid, callback_url, callback_method)

    @validate_args(call_uuid=[of_type(six.text_type)],
                   callback_url=[optional(is_url())],
                   callback_method=[optional(of_type(six.text_type))]
                   )
    def stop_playing(self, call_uuid, callback_url=None, callback_method=None):
        return self.client.request('DELETE', ('Call', call_uuid, 'Play'), to_param_dict(self.stop_playing, locals()),
                                   is_voice_request=True)

    @validate_args(call_uuid=[of_type(six.text_type)],
                   callback_url=[optional(is_url())],
                   callback_method=[optional(of_type(six.text_type))],
                   )
    def speak(self,
              call_uuid,
              text,
              voice=None,
              language=None,
              legs=None,
              loop=None,
              mix=None,
              callback_url=None,
              callback_method=None,
              type=None,
              ):
        return self.start_speaking(call_uuid, text, voice, language, legs,
                                   loop, mix, callback_url, callback_method, type)

    @validate_args(call_uuid=[of_type(six.text_type)])
    def start_speaking(self,
                       call_uuid,
                       text,
                       voice=None,
                       language=None,
                       legs=None,
                       loop=None,
                       mix=None,
                       callback_url=None,
                       callback_method=None,
                       type=None
                       ):
        return self.client.request('POST', ('Call', call_uuid, 'Speak'),
                                   to_param_dict(self.start_speaking,
                                                 locals()), is_voice_request=True)

    @validate_args(call_uuid=[of_type(six.text_type)])
    def stop_speaking(self, call_uuid, callback_url=None, callback_method=None):
        return self.client.request('DELETE', ('Call', call_uuid, 'Speak'), to_param_dict(self.stop_speaking, locals()),
                                   is_voice_request=True)

    @validate_args(
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))]
    )
    def speak_stop(self, call_uuid, callback_url=None, callback_method=None):
        return self.client.calls.stop_speaking(call_uuid, callback_url, callback_method)

    @validate_args(call_uuid=[of_type(six.text_type)],
                   callback_url=[optional(is_url())],
                   callback_method=[optional(of_type(six.text_type))],
                   )
    def send_digits(self, call_uuid, digits, leg=None, callback_url=None, callback_method=None):
        return self.client.request('POST', ('Call', call_uuid, 'DTMF'),
                                   to_param_dict(self.send_digits, locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        callback_url=[optional(is_url())],
        callback_method=[optional(of_type(six.text_type))],
    )
    def delete(self,
               call_uuid,
               callback_url=None,
               callback_method=None):
        return self.client.request('DELETE', ('Call', call_uuid), to_param_dict(self.delete, locals()),
                                   is_voice_request=True)

    @validate_args(call_uuid=[of_type(six.text_type)])
    def hangup(self, call_uuid):
        return self.client.calls.delete(call_uuid)

    def cancel(self, request_uuid):
        return self.client.request('DELETE', ('Request', request_uuid), is_voice_request=True)

    def live_call_list_ids(self, limit=None, offset=None):
        return self.client.live_calls.list_ids(limit, offset)

    def live_call_get(self, _id):
        return self.client.live_calls.get(_id)

    def queued_call_list_ids(self, limit=None, offset=None):
        return self.client.queued_calls.list_ids(limit, offset)

    def queued_call_get(self, _id):
        return self.client.queued_calls.get(_id)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        service_url=[of_type(six.text_type)]
    )
    def start_stream(self,
                     call_uuid,
                     service_url,
                     bidirectional=None,
                     audio_track=None,
                     stream_timeout=None,
                     status_callback_url=None,
                     status_callback_method=None,
                     content_type=None,
                     extra_headers=None):
        return self.client.request('POST', ('Call', call_uuid, 'Stream'),
                                   to_param_dict(self.start_stream,
                                                 locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        stream_id=[of_type(six.text_type)]
    )
    def delete_specific_stream(self,
                               call_uuid,
                               stream_id):
        return self.client.request('DELETE', ('Call', call_uuid, 'Stream', stream_id),
                                   to_param_dict(self.delete_specific_stream,
                                                 locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)]
    )
    def delete_all_streams(self,
                           call_uuid):
        return self.client.request('DELETE', ('Call', call_uuid, 'Stream'),
                                   to_param_dict(self.delete_all_streams,
                                                 locals()), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)],
        stream_id=[of_type(six.text_type)]
    )
    def get_details_of_specific_stream(self,
                                       call_uuid,
                                       stream_id):
        return self.client.request('GET', ('Call', call_uuid, 'Stream', stream_id), is_voice_request=True)

    @validate_args(
        call_uuid=[of_type(six.text_type)]
    )
    def get_all_streams(self,
                        call_uuid):
        return self.client.request('GET', ('Call', call_uuid, 'Stream'), is_voice_request=True)




from plivo.utils.validators import *
from plivo.xml import (PlivoXMLElement,
                       map_type,
                       BreakElement,
                       )


class LangElement(PlivoXMLElement):
    _name = 'lang'
    _nestable = [
        "break",
        "emphasis",
        "lang",
        "p",
        "phoneme",
        "prosody",
        "s",
        "say-as",
        "sub",
        "w"
    ]

    @property
    def xmllang(self):
        return self.__xmllang

    @xmllang.setter
    def xmllang(self, value):
        self.__xmllang = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_xmllang(self, value):
        self.xmllang = value
        return self

    def __init__(
            self,
            content=None,
            xmllang=None,
    ):

        super(LangElement, self).__init__()
        self.content = content
        self.xmllang = xmllang

    def to_dict(self):
        d = {
            '{http://www.w3.org/XML/1998/namespace}lang': self.xmllang,
        }

        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
            }

    def add_break(
            self,
            strength=None,
            time=None
    ):
        self.add(
            BreakElement(
                strength=strength,
                time=time,
            ))
        return self

    def add_emphasis(
            self,
            content,
            level=None,
    ):
        from .emphasisElement import EmphasisElement

        self.add(
            EmphasisElement(
                content=content,
                level=level,
            ))
        return self

    def add_lang(
            self,
            content,
            xmllang=None,
    ):
        self.add(
            LangElement(
                content=content,
                xmllang=xmllang,
            ))
        return self

    def add_p(
            self,
            content,
    ):
        from .pElement import PElement

        self.add(
            PElement(
                content=content,
            ))
        return self

    def add_phoneme(
            self,
            content,
            alphabet=None,
            ph=None,
    ):
        from .phonemeElement import PhonemeElement

        self.add(
            PhonemeElement(
                content=content,
                alphabet=alphabet,
                ph=ph,
            ))
        return self

    def add_prosody(
            self,
            content,
            volume=None,
            rate=None,
            pitch=None,
    ):
        from .prosodyElement import ProsodyElement

        self.add(
            ProsodyElement(
                content=content,
                volume=volume,
                rate=rate,
                pitch=pitch,
            ))
        return self

    def add_s(
            self,
            content,
    ):
        from .sElement import SElement

        self.add(
            SElement(
                content=content,
            ))
        return self

    def add_say_as(
            self,
            content,
            interpret_as=None,
            format=None,
    ):
        from .sayAsElement import SayAsElement

        self.add(
            SayAsElement(
                content=content,
                interpret_as=interpret_as,
                format=format,
            ))
        return self

    def add_sub(
            self,
            content,
            alias=None,
    ):
        from .subElement import SubElement

        self.add(
            SubElement(
                content=content,
                alias=alias,
            ))
        return self

    def add_w(
            self,
            content,
            role=None,
    ):
        from .wElement import WElement

        self.add(
            WElement(
                content=content,
                role=role,
            ))
        return self



# -*- coding: utf-8 -*-
import functools
import importlib
import inspect
import re

import decorator
import six

from plivo.exceptions import ValidationError


def regex(regex):
    rexp = re.compile(regex)

    def f(name, value):
        if not rexp.match(value):
            return None, [
                '{} should match format {} (actual value: {})'.format(
                    name, regex, value)
            ]

        return value, []

    return f


def all_of(*validators):
    def f(name, value):
        for validator in validators:
            value, errs = validator(name, value)
            if errs:
                return None, errs
        return value, []

    return f


def one_of(*validators):
    def f(name, value):
        for validator in validators:
            new_val, errs = validator(name, value)
            if errs:
                continue
            else:
                return new_val, []
        return None, '{} did not satisfy any of the required types'.format(
            name)

    return f


def check(checker, message=None):
    def f(name, value):
        inr = checker(value)
        msg = message or '{} should be in range'.format(name)
        if inr:
            return value, []
        else:
            return None, ['{} (actual value: {})'.format(msg, value)]

    return f


def is_in(iterable, message=None, case_sensitive=True, case_type='upper'):
    def f(name, value):
        actual_value = value
        if not case_sensitive:
            if case_type == 'upper':
                value = str(value).upper()
            elif case_type == 'lower':
                value = str(value).lower()
            elif case_type == 'title':
                value = str(value).title()

        msg = message or '{} should be in {}'.format(name, iterable)
        if value in iterable:
            return value, []
        else:
            return None, ['{} (actual value: {})'.format(msg, actual_value)]
    return f


def multi_is_in(iterable, message=None, case_sensitive=True, make_lower_case=False, separator=','):
    def f(name, value):
        actual_value = value
        if not case_sensitive:
            if make_lower_case:
                value = str(value).lower()
            else:
                value = str(value).upper()
        msg = message or '{} should be among {}. multiple values should be COMMA(,) separated'.format(name, iterable)
        for val in value.split(separator):
            if val not in iterable:
                return None, ['{} (actual value: {})'.format(msg, actual_value)]
        return value, []

    return f


def optional(*validators):
    def f(name, value):
        if value is None:
            return None, []
        return all_of(*validators)(name, value)

    return f


def required(validate):
    def f(name, value):
        if not value:
            return None, '{} is required (current value: None)'.format(name)
        return validate(name, value)

    return f


def of_type(*args):
    def f(name, value):
        if value is None:
            return None, ['{name} cannot be None'.format(name=name)]
        for typ in args:
            if isinstance(typ, str):
                parts = typ.split('.')
                typ = getattr(
                    importlib.import_module('.'.join(parts[:-1])), parts[-1])
            try:
                value = typ(value)
                return value, []
            except ValueError:
                pass
        return None, [
            '{name} should be of type: {types}'.format(
                name=name, types=[arg.__name__ for arg in args])
        ]

    return f


def of_type_exact(*args):
    def f(name, value):
        for typ in args:
            if isinstance(typ, str):
                parts = typ.split('.')
                typ = getattr(
                    importlib.import_module('.'.join(parts[:-1])), parts[-1])
            if not isinstance(value, typ):
                continue
            return value, []
        return None, [
            '{name} should be of type: {types}'.format(
                name=name,
                types=[getattr(arg, '__name__', str(arg)) for arg in args])
        ]

    return f



def is_iterable(validator, sep=None):
    def f(name, value):
        try:
            l = []
            v, e = validator(name, value)
            if v and not e:
                raise TypeError()  # hack
            for i, item in enumerate(iter(value)):
                val, errs = validator('{}[{}]'.format(name, i), item)
                if errs:
                    return None, errs
                l.append(val)
            ret = (l, []) if not sep else (sep.join(l), [])
            return ret
        except TypeError:
            val, errs = validator(name, value)
            if errs:
                return None, errs
            return ([val], []) if not sep else (val, [])

    return required(f)


def multiple_valid_integers():
    def f(name, value):
        if isinstance(value, str):
            values = value.split('<')
            for i in values:
                try:
                    int(i)
                except ValueError:
                    return None, ['{} destination value must be integer'.format(name)]
            return value, []
        return value, []
    return f


def validate_args(**to_validate):
    def outer(wrapped):
        @functools.wraps(wrapped)
        def wrapper(self, *args, **kwargs):
            params = inspect.getcallargs(wrapped, *args, **kwargs)
            for arg_name, validators in to_validate.items():
                for validator in validators:
                    params[arg_name], errs = validator(arg_name,
                                                       params.get(
                                                           arg_name, None))
                    if errs:
                        raise ValidationError(errs)
            return wrapped(**params)

        return decorator.decorate(wrapped, wrapper)

    return outer

def validate_list_items(instance_type):
    def f(arg_name, value):
        if not isinstance(value, list):
            return [], ["{} must be a list".format(arg_name)]

        errors = []
        validated_items = []
        for idx, item in enumerate(value):
            flag = 0
            if isinstance(item, dict):
                try:
                    instance_type(**item)
                except Exception as exception:
                    errors.append(str(exception))
                    flag=1

            else:
                err = "Invalid item at index {} in {}: should be of type {}".format(idx, arg_name, instance_type.__name__)
                if not isinstance(item, instance_type):
                    errors.append(err)
                    flag = 1
                
            if not flag:
                validated_items.append(item)

        return validated_items, errors

    return f

is_valid_date = functools.partial(of_type, six.text_type)
is_phonenumber = functools.partial(of_type, six.text_type)
is_subaccount_id = functools.partial(all_of, of_type(six.text_type),
                                     regex(r'^SA[A-Z0-9]{18}$'))
is_mainaccount_id = functools.partial(all_of, of_type(six.text_type),
                                      regex(r'^MA[A-Z0-9]{18}$'))
is_account_id = functools.partial(all_of, of_type(six.text_type),
                                      regex(r'^(M|S)A[A-Z0-9]{18}$'))
is_subaccount = functools.partial(
    one_of, of_type_exact('plivo.resources.accounts.Subaccount'),
    is_subaccount_id())
is_url = functools.partial(
    all_of, of_type(six.text_type),
    regex(
        r'(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+|None)'
    ))
is_proper_date_format = functools.partial(all_of, of_type_exact(str),
                                          regex(r'^\d{4}-\d{2}-\d{2} \d{2}:\d{2}(:\d{2}(\.\d{1,6})?)?$'))

is_template = functools.partial(of_type_exact,'plivo.utils.template.Template')



from plivo.base import (ListResponseObject, PlivoResource,
                        PlivoResourceInterface)
from plivo.utils import to_param_dict


class VerifyCallerid(PlivoResource):
    _name = 'VerifyCallerid'

    def initiate_verify(self):
        return self.client.calls.initiate_verify(self.id,
                                                 **to_param_dict(self.initiate_verify(), locals()))

    def verify_caller_id(self):
        return self.client.calls.verify_caller_id(self.id,
                                                  **to_param_dict(self.verify_caller_id(), locals()))

    def delete_verified_caller_id(self):
        return self.client.calls.delete_verified_caller_id(self.id,
                                                           **to_param_dict(self.delete_verified_caller_id(), locals()))

    def get_verified_caller_id(self):
        return self.client.calls.get_verified_caller_id(self.id,
                                                        **to_param_dict(self.get_verified_caller_id(), locals()))

    def update_verified_caller_id(self):
        return self.client.calls.update_verified_caller_id(self.id,
                                                           **to_param_dict(self.update_verified_caller_id(), locals()))

    def list_verified_caller_id(self):
        return self.client.calls.list_verified_caller_id(self.id,
                                                         **to_param_dict(self.list_verified_caller_id(), locals()))


class VerifyCallerids(PlivoResourceInterface):
    _resource_type = VerifyCallerid

    def initiate_verify(self,
                        phone_number=None,
                        alias=None,
                        channel=None,
                        country=None,
                        subaccount=None,
                        account_id=None,
                        auth_id=None,
                        auth_token=None
                        ):
        return self.client.request('POST', ('VerifiedCallerId',),
                                   to_param_dict(self.initiate_verify, locals()), is_voice_request=True)

    def verify_caller_id(self,
                         verification_uuid,
                         otp=None,
                         ):
        return self.client.request('POST', ('VerifiedCallerId', 'Verification', verification_uuid),
                                   to_param_dict(self.verify_caller_id, locals()), is_voice_request=True)

    def delete_verified_caller_id(self,
                                  phone_number
                                  ):
        return self.client.request('DELETE', ('VerifiedCallerId', phone_number),
                                   to_param_dict(self.delete_verified_caller_id, locals()), is_voice_request=True)

    def get_verified_caller_id(self,
                               phone_number
                               ):
        return self.client.request('GET', ('VerifiedCallerId', phone_number), is_voice_request=True)

    def update_verified_caller_id(self,
                                  phone_number,
                                  alias=None,
                                  subaccount=None
                                  ):
        return self.client.request('POST', ('VerifiedCallerId', phone_number),
                                   to_param_dict(self.update_verified_caller_id, locals()), is_voice_request=True)

    def list_verified_caller_id(self,
                                country=None,
                                subaccount=None,
                                alias=None,
                                limit=None,
                                offset=None
                                ):
        return self.client.request('GET', ('VerifiedCallerId',), to_param_dict(self.list_verified_caller_id, locals()),
                                   is_voice_request=True)



from plivo.utils.validators import *

class Parameter:
    @validate_args(
        type=[required(of_type_exact(str))],
        text=[optional(of_type_exact(str, type(None)))],
        media=[optional(of_type_exact(str))],
        currency=[optional(of_type_exact(dict))],
        date_time=[optional(of_type_exact(dict))],
    )
    def __init__(self, type, text=None, media=None, currency=None, date_time=None):
        self.type = type
        self.text = text
        self.media = media
        self.currency = Currency(**currency) if currency else None
        self.date_time = DateTime(**date_time) if date_time else None

class Component:
    @validate_args(
        type=[required(of_type_exact(str))],
        sub_type=[optional(of_type_exact(str, type(None)))],
        index=[optional(of_type_exact(str, type(None)))],
        parameters=[optional(validate_list_items(Parameter))],
    )
    def __init__(self, type, sub_type=None, index=None, parameters=None):
        self.type = type
        self.sub_type = sub_type
        self.index = index
        self.parameters = parameters if parameters is not None else []

class Template:
    @validate_args(
        name=[required(of_type_exact(str))],
        language=[required(of_type_exact(str))],
        components=[optional(validate_list_items(Component))],
    )
    def __init__(self, name, language, components=None):
        self.name = name
        self.language = language
        self.components = components if components is not None else []


class Currency:
    @validate_args(
        fallback_value=[required(of_type_exact(str))],
        currency_code=[required(of_type_exact(str))],
        amount_1000=[required(of_type_exact(int))],
    )
    def __init__(self, fallback_value, currency_code, amount_1000):
        self.fallback_value=fallback_value
        self.currency_code = currency_code
        self.amount_1000 = amount_1000


class DateTime:
    @validate_args(
        fallback_value=[required(of_type_exact(str))],
    )
    def __init__(self, fallback_value):
        self.fallback_value = fallback_value



from plivo.utils.validators import *
from plivo.xml import PlivoXMLElement, map_type


class ConferenceElement(PlivoXMLElement):
    _name = 'Conference'
    _nestable = []

    @property
    def muted(self):
        return self.__muted

    @muted.setter
    def muted(self, value):
        self.__muted = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_muted(self, value):
        self.muted = value
        return self

    @property
    def enter_sound(self):
        return self.__enter_sound

    @enter_sound.setter
    def enter_sound(self, value):
        self.__enter_sound = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_enter_sound(self, value):
        self.enter_sound = value
        return self

    @property
    def exit_sound(self):
        return self.__exit_sound

    @exit_sound.setter
    def exit_sound(self, value):
        self.__exit_sound = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_exit_sound(self, value):
        self.exit_sound = value
        return self

    @property
    def start_conference_on_enter(self):
        return self.__start_conference_on_enter

    @start_conference_on_enter.setter
    def start_conference_on_enter(self, value):
        self.__start_conference_on_enter = bool(
            value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_start_conference_on_enter(self, value):
        self.start_conference_on_enter = value
        return self

    @property
    def end_conference_on_exit(self):
        return self.__end_conference_on_exit

    @end_conference_on_exit.setter
    def end_conference_on_exit(self, value):
        self.__end_conference_on_exit = bool(
            value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_end_conference_on_exit(self, value):
        self.end_conference_on_exit = value
        return self

    @property
    def stay_alone(self):
        return self.__stay_alone

    @stay_alone.setter
    def stay_alone(self, value):
        self.__stay_alone = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_stay_alone(self, value):
        self.stay_alone = value
        return self

    @property
    def wait_sound(self):
        return self.__wait_sound

    @wait_sound.setter
    def wait_sound(self, value):
        self.__wait_sound = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_wait_sound(self, value):
        self.wait_sound = value
        return self

    @property
    def max_members(self):
        return self.__max_members

    @max_members.setter
    def max_members(self, value):
        self.__max_members = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_max_members(self, value):
        self.max_members = value
        return self

    @property
    def record(self):
        return self.__record

    @record.setter
    def record(self, value):
        self.__record = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_record(self, value):
        self.record = value
        return self

    @property
    def record_file_format(self):
        return self.__record_file_format

    @record_file_format.setter
    def record_file_format(self, value):
        self.__record_file_format = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_record_file_format(self, value):
        self.record_file_format = value
        return self

    @property
    def time_limit(self):
        return self.__time_limit

    @time_limit.setter
    def time_limit(self, value):
        self.__time_limit = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_time_limit(self, value):
        self.time_limit = value
        return self

    @property
    def hangup_on_star(self):
        return self.__hangup_on_star

    @hangup_on_star.setter
    def hangup_on_star(self, value):
        self.__hangup_on_star = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_hangup_on_star(self, value):
        self.hangup_on_star = value
        return self

    @property
    def action(self):
        return self.__action

    @action.setter
    def action(self, value):
        self.__action = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_action(self, value):
        self.action = value
        return self

    @property
    def method(self):
        return self.__method

    @method.setter
    def method(self, value):
        self.__method = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_method(self, value):
        self.method = value
        return self

    @property
    def callback_url(self):
        return self.__callback_url

    @callback_url.setter
    def callback_url(self, value):
        self.__callback_url = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_callback_url(self, value):
        self.callback_url = value
        return self

    @property
    def callback_method(self):
        return self.__callback_method

    @callback_method.setter
    def callback_method(self, value):
        self.__callback_method = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_callback_method(self, value):
        self.callback_method = value
        return self

    @property
    def digits_match(self):
        return self.__digits_match

    @digits_match.setter
    def digits_match(self, value):
        self.__digits_match = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_digits_match(self, value):
        self.digits_match = value
        return self

    @property
    def floor_event(self):
        return self.__floor_event

    @floor_event.setter
    def floor_event(self, value):
        self.__floor_event = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_floor_event(self, value):
        self.floor_event = value
        return self

    @property
    def redirect(self):
        return self.__redirect

    @redirect.setter
    def redirect(self, value):
        self.__redirect = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_redirect(self, value):
        self.redirect = value
        return self

    @property
    def relay_dtmf(self):
        return self.__relay_dtmf

    @relay_dtmf.setter
    def relay_dtmf(self, value):
        self.__relay_dtmf = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_relay_dtmf(self, value):
        self.relay_dtmf = value
        return self

    @property
    def send_digits(self):
        return self.__send_digits

    @send_digits.setter
    def send_digits(self, value):
        self.__send_digits = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_send_digits(self, value):
        self.send_digits = value
        return self

    @property
    def record_when_alone(self):
        return self.__record_when_alone

    @record_when_alone.setter
    def record_when_alone(self, value):
        self.__record_when_alone = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_record_when_alone(self, value):
        self.record_when_alone = value
        return self

    @property
    def transcription_type(self):
        return self.__transcription_type

    @transcription_type.setter
    def transcription_type(self, value):
        self.__transcription_type = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_transcription_type(self, value):
        self.transcription_type = value
        return self

    @property
    def transcription_url(self):
        return self.__transcription_url

    @transcription_url.setter
    def transcription_url(self, value):
        self.__transcription_url = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_transcription_url(self, value):
        self.transcription_url = value
        return self

    @property
    def transcription_method(self):
        return self.__transcription_method

    @transcription_method.setter
    def transcription_method(self, value):
        self.__transcription_method = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_transcription_method(self, value):
        self.transcription_method = value
        return self

    def __init__(
            self,
            content=None,
            muted=None,
            enter_sound=None,
            exit_sound=None,
            start_conference_on_enter=None,
            end_conference_on_exit=None,
            stay_alone=None,
            wait_sound=None,
            max_members=None,
            record=None,
            record_file_format=None,
            time_limit=None,
            hangup_on_star=None,
            action=None,
            method=None,
            callback_url=None,
            callback_method=None,
            digits_match=None,
            floor_event=None,
            redirect=None,
            relay_dtmf=None,
            send_digits=None,
            record_when_alone=None,
            transcription_type=None,
            transcription_url=None,
            transcription_method=None,
    ):
        super(ConferenceElement, self).__init__()

        self.content = content
        self.muted = muted
        self.enter_sound = enter_sound
        self.exit_sound = exit_sound
        self.start_conference_on_enter = start_conference_on_enter
        self.end_conference_on_exit = end_conference_on_exit
        self.stay_alone = stay_alone
        self.wait_sound = wait_sound
        self.max_members = max_members
        self.record = record
        self.record_file_format = record_file_format
        self.time_limit = time_limit
        self.hangup_on_star = hangup_on_star
        self.action = action
        self.method = method
        self.callback_url = callback_url
        self.callback_method = callback_method
        self.digits_match = digits_match
        self.floor_event = floor_event
        self.redirect = redirect
        self.relay_dtmf = relay_dtmf
        self.send_digits = send_digits
        self.record_when_alone = record_when_alone
        self.transcription_type = transcription_type
        self.transcription_url = transcription_url
        self.transcription_method = transcription_method

    def to_dict(self):
        d = {
            'muted': self.muted,
            'enterSound': self.enter_sound,
            'exitSound': self.exit_sound,
            'startConferenceOnEnter': self.start_conference_on_enter,
            'endConferenceOnExit': self.end_conference_on_exit,
            'stayAlone': self.stay_alone,
            'waitSound': self.wait_sound,
            'maxMembers': self.max_members,
            'record': self.record,
            'recordFileFormat': self.record_file_format,
            'timeLimit': self.time_limit,
            'hangupOnStar': self.hangup_on_star,
            'action': self.action,
            'method': self.method,
            'callbackUrl': self.callback_url,
            'callbackMethod': self.callback_method,
            'digitsMatch': self.digits_match,
            'floorEvent': self.floor_event,
            'redirect': self.redirect,
            'relayDTMF': self.relay_dtmf,
            'sendDigits': self.send_digits,
            'recordWhenAlone': self.record_when_alone,
            'transcriptionType': self.transcription_type,
            'transcriptionUrl': self.transcription_url,
            'transcriptionMethod': self.transcription_method,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



from .xmlUtils import map_type
from .PlivoXMLElement import PlivoXMLElement
from .ConferenceElement import ConferenceElement
from .DTMFElement import DTMFElement
from .MultiPartyCallElement import MultiPartyCallElement
from .hangupElement import HangupElement
from .messageElement import MessageElement
from .numberElement import NumberElement
from .playElement import PlayElement
from .waitElement import WaitElement
from .recordElement import RecordElement
from .redirectElement import RedirectElement
from .userElement import UserElement
from .breakElement import BreakElement
from .langElement import LangElement
from .emphasisElement import EmphasisElement
from .speakElement import SpeakElement
from .getDigitsElement import GetDigitsElement
from .getInputElement import GetInputElement
from .preAnswerElement import PreAnswerElement
from .DialElement import DialElement
from .ResponseElement import ResponseElement
from .pElement import PElement
from .phonemeElement import PhonemeElement
from .prosodyElement import ProsodyElement
from .sElement import SElement
from .sayAsElement import SayAsElement
from .subElement import SubElement
from .wElement import WElement
from .contElement import ContElement
from .streamElement import StreamElement


from plivo.xml import PlivoXMLElement, map_type
from plivo.utils.validators import *


class HangupElement(PlivoXMLElement):
    _name = 'Hangup'
    _nestable = []

    @property
    def reason(self):
        return self.__reason

    @reason.setter
    def reason(self, value):
        self.__reason = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_reason(self, value):
        self.reason = value
        return self

    @property
    def schedule(self):
        return self.__schedule

    @schedule.setter
    def schedule(self, value):
        self.__schedule = int(value) if value is not None else None

    @validate_args(
        value=[of_type(*six.integer_types)],
    )
    def set_schedule(self, value):
        self.schedule = value
        return self

    def __init__(
            self,
            reason=None,
            schedule=None,
    ):
        super(HangupElement, self).__init__()

        self.reason = reason
        self.schedule = schedule

    def to_dict(self):
        d = {
            'reason': self.reason,
            'schedule': self.schedule,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



# -*- coding: utf-8 -*-
class PlivoRestError(Exception):
    pass


class AuthenticationError(PlivoRestError):
    pass


class InvalidRequestError(PlivoRestError):
    pass


class PlivoServerError(PlivoRestError):
    pass


class PlivoXMLError(PlivoRestError):
    pass


class ResourceNotFoundError(PlivoRestError):
    pass


class ValidationError(PlivoRestError):
    pass


class ForbiddenError(PlivoRestError):
    pass



import six

from plivo.xml import (
    PlayElement,
    PlivoXMLElement,
    SpeakElement,
    WaitElement,
    map_type,
    DTMFElement,
    GetDigitsElement,
    RedirectElement,
    MessageElement,
)


class PreAnswerElement(PlivoXMLElement):
    _name = 'PreAnswer'
    _nestable = [
        'Speak',
        'Play',
        'Wait',
        'GetDigits',
        'Redirect',
        'Message',
        'DTMF'
    ]

    def __init__(self):
        super(PreAnswerElement, self).__init__()

    def to_dict(self):
        d = {}
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }

    def add_speak(
            self,
            content,
            voice=None,
            language=None,
            loop=None, ):
        self.add(
            SpeakElement(
                content=content,
                voice=voice,
                language=language,
                loop=loop, ))
        return self

    def add_play(
            self,
            content,
            loop=None, ):
        self.add(PlayElement(
            content=content,
            loop=loop, ))
        return self

    def add_wait(
            self,
            length=None,
            silence=None,
            min_silence=None,
            beep=None, ):
        self.add(
            WaitElement(
                length=length,
                silence=silence,
                min_silence=min_silence,
                beep=beep, ))
        return self

    def add_get_digits(
            self,
            action=None,
            method=None,
            timeout=None,
            digit_timeout=None,
            finish_on_key=None,
            num_digits=None,
            retries=None,
            redirect=None,
            play_beep=None,
            valid_digits=None,
            invalid_digits_sound=None,
            log=None,
    ):
        self.add(
            GetDigitsElement(
                action=action,
                method=method,
                timeout=timeout,
                digit_timeout=digit_timeout,
                finish_on_key=finish_on_key,
                num_digits=num_digits,
                retries=retries,
                redirect=redirect,
                play_beep=play_beep,
                valid_digits=valid_digits,
                invalid_digits_sound=invalid_digits_sound,
                log=log,
            ))
        return self

    def add_redirect(
            self,
            content,
            method=None,
    ):
        self.add(
            RedirectElement(
                content=content,
                method=method,
            ))
        return self

    def add_message(
            self,
            content,
            src=None,
            dst=None,
            type=None,
            callback_url=None,
            callback_method=None,
    ):
        self.add(
            MessageElement(
                content=content,
                src=src,
                dst=dst,
                type=type,
                callback_url=callback_url,
                callback_method=callback_method,
            ))
        return self

    def add_dtmf(
            self,
            content,
            async_=None,
    ):
        self.add(
            DTMFElement(
                content=content,
                async_=async_,
            ))
        return self



from plivo.utils.validators import *
from plivo.xml import PlivoXMLElement, map_type


class SubElement(PlivoXMLElement):
    _name = 'sub'
    _nestable = []

    @property
    def alias(self):
        return self.__alias

    @alias.setter
    def alias(self, value):
        self.__alias = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_alias(self, value):
        self.alias = value
        return self

    def __init__(
        self,
        content=None,
        alias=None,
    ):
        super(SubElement, self).__init__()

        self.content = content
        self.alias = alias

    def to_dict(self):
        d = {
            'alias': self.alias,
        }

        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



from plivo.xml import PlivoXMLElement, map_type
from plivo.utils.validators import *


class NumberElement(PlivoXMLElement):
    _name = 'Number'
    _nestable = []

    @property
    def send_digits(self):
        return self.__send_digits

    @send_digits.setter
    def send_digits(self, value):
        self.__send_digits = six.text_type(
            value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_send_digits(self, value):
        self.send_digits = value
        return self

    @property
    def send_on_preanswer(self):
        return self.__send_on_preanswer

    @send_on_preanswer.setter
    def send_on_preanswer(self, value):
        self.__send_on_preanswer = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_send_on_preanswer(self, value):
        self.send_on_preanswer = value
        return self

    @property
    def send_digits_mode(self):
        return self.__send_digits_mode

    @send_digits_mode.setter
    def send_digits_mode(self, value):
        self.__send_digits_mode = six.text_type(value) if value is not None else None

    @validate_args(
        value=[of_type(six.text_type)],
    )
    def set_send_digits_mode(self, value):
        self.send_digits_mode = value
        return self

    def __init__(
            self,
            content,
            send_digits=None,
            send_on_preanswer=None,
            send_digits_mode=None):
        super(NumberElement, self).__init__()

        self.content = str(content)
        self.send_digits = send_digits
        self.send_on_preanswer = send_on_preanswer
        self.send_digits_mode = send_digits_mode

    def to_dict(self):
        d = {
            'sendDigits': self.send_digits,
            'sendOnPreanswer': self.send_on_preanswer,
            'sendDigitsMode': self.send_digits_mode,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }



from plivo.xml import PlivoXMLElement, map_type
from plivo.utils.validators import *


class DTMFElement(PlivoXMLElement):
    _name = 'DTMF'
    _nestable = []

    @property
    def async_(self):
        return self.__async

    @async_.setter
    def async_(self, value):
        self.__async = bool(value) if value is not None else None

    @validate_args(
        value=[of_type_exact(bool)],
    )
    def set_async(self, value):
        self.async_ = value
        return self

    def __init__(
            self,
            content,
            async_=None,
    ):
        super(DTMFElement, self).__init__()

        self.content = content
        self.async_ = async_

    def to_dict(self):
        d = {
            'async': self.async_,
        }
        return {
            k: six.text_type(map_type(v))
            for k, v in d.items() if v is not None
        }
